#define MY_RADIO_NRF24
#define MY_DEBUG 

#pragma region Includes
//#include <Adafruit_SleepyDog.h>  //Werkt niet wegens interferentie van __vector_6
#include <SPI.h>
#include <MySensors.h>
#include <SparkFunTSL2561.h>
#include <Wire.h> 
#include <HIH6130.h>
#include <Adafruit_INA219.h>
#pragma endregion

#pragma region Child id definitions
#define CHILD_ID_LIGHT 1
#define CHILD_ID_MULTIMETER 2
#define CHILD_ID_TEMPERATURE 3
#define CHILD_ID_HUMIDITY 4
#define CHILD_ID_DOOR 5
#define CHILD_ID_MOTORCURRENT 7
#define CHILD_ID_RELAIS1 8
#define CHILD_ID_RELAIS2 9
//#define CHILD_ID_POWER 10
#pragma endregion

#pragma region Motor timing parameter settings
#define MAX_LOOPTIJD 6000	
#define EXTRA_TIME_CLOSE 100	
#define RETURN_AFTER_FAULT_TIME 2000	//was 2000
#define CAMERA_ON_TIME 90 // in seconds  was 60
#define DOOR_AFTER_CAMERA_TIME 60 // in seconds was 30
#define SLEEP_CYCLES 100
#define SLEEP_PERIOD 200
#define LUX_CLOSE 7.0F
#define LUX_OPEN 13.0F
#pragma endregion

#pragma region Pin definitions
//Pin definitions
#define PushButtonOpenPin 3 // Gele kabel (rechts) sluiten op wit (midden)
#define PushbuttonClosePin A3 // Blauwe kabel (links) sluiten op wit (midden)
#define RelayMotorOpenPin 5
#define RelayMotorClosePin 4
#define Relay2Pin 6
#define Relay1Pin 7
#define PirPin 8
#define EndDetectionOpenPin A1  //Zwarte kabel (rechts) sluiten op wit (midden)
#define EndDetectionClosePin A0  //Groene kabel (Links) sluiten op wit (midden)
#define BatteryVoltagePin A2
#define MotorStatus_Stop 0
#define MotorStatus_Open 1
#define MotorStatus_Close 2
#define MotorStatus_Retour 3
#define DoorStatus_InBetween 0
#define DoorStatus_Open 1
#define DoorStatus_Closed 2
#pragma endregion

#pragma region Library object declarations
byte HIH6130_address = 0x27;
HIH6130 rht(HIH6130_address);
Adafruit_INA219 ina;
SFE_TSL2561 light;
#pragma endregion

#pragma region MySensors message declarations
MyMessage lightMsg(CHILD_ID_LIGHT, V_LIGHT_LEVEL);
MyMessage currentMsg(CHILD_ID_MULTIMETER, V_CURRENT);
MyMessage voltageMsg(CHILD_ID_MULTIMETER, V_VOLTAGE);
MyMessage temperatureMsg(CHILD_ID_MULTIMETER, V_TEMP);
MyMessage humidityMsg(CHILD_ID_MULTIMETER, V_HUM);
MyMessage doorStatusMsg(CHILD_ID_DOOR, V_LOCK_STATUS);
MyMessage motorCurrentMsg(CHILD_ID_MOTORCURRENT, V_CURRENT);
MyMessage relais1Msg(CHILD_ID_RELAIS1, V_LIGHT);
//MyMessage powerWattMsg(CHILD_ID_POWER, V_WATT);
//MyMessage powerkWhMsg(CHILD_ID_POWER, V_KWH);
#pragma endregion

#pragma region Global status variables
byte motorStatus;
byte doorStatus;
byte lastDoorStatus = DoorStatus_InBetween;
bool buttonOpenLastStatus = false;
bool buttonCloseLastStatus = false;
bool doorOpenRequested = false;
bool doorCloseRequested = false;
bool cameraOn = false;
bool cameraActivatedByLuxSensor = false;
uint32_t cameraStartTime = 0;
bool buttonOpenPressed = false;
bool buttonClosePressed = false;
unsigned long motorStopTime;
boolean gain;     // Gain setting, 0 = X1, 1 = X16;
unsigned char time = 0;
unsigned int ms;  // Integration ("shutter") time in milliseconds
#pragma endregion

#pragma region Global variables for sensor measurements
double lux;
float temp;
float humidity;
float current;
float motorCurrent;
float currentSum;
int currentCounter;
float voltage;
#pragma endregion


void setup()  
{ 
	Serial.begin(115200);
	Serial.println("Kippenhok_4");

	#pragma region Pin setup and initialization
	pinMode(PushButtonOpenPin, INPUT_PULLUP);
	pinMode(PushbuttonClosePin, INPUT_PULLUP);
	pinMode(RelayMotorOpenPin, OUTPUT);
	pinMode(RelayMotorClosePin, OUTPUT);
	pinMode(Relay1Pin, OUTPUT);
	pinMode(Relay2Pin, OUTPUT);
	digitalWrite(RelayMotorOpenPin, LOW);
	digitalWrite(RelayMotorClosePin, LOW);
	digitalWrite(Relay1Pin, LOW);
	digitalWrite(Relay2Pin, LOW);
	pinMode(PirPin, INPUT_PULLUP);
	pinMode(EndDetectionOpenPin, INPUT_PULLUP);
	pinMode(EndDetectionClosePin, INPUT_PULLUP);
	#pragma endregion

	ActivateMotor(MotorStatus_Stop);

	#pragma region Library object initialization
	rht.begin();
	ina.begin();
	light.begin();
	unsigned char ID;
	gain = 0;
	time = 1;
	light.setTiming(gain, time, ms);
	light.setPowerUp();
	delay(500);
	Serial.println("application started");
	requestTime();
}

uint16_t sleepCounter = 0;
int luxState = 0;  // = unknown
uint32_t controllerTime;
uint32_t secondsAfterMidnight=0;
byte hours = 0;
byte minutes = 0;
byte seconds = 0;
bool timeReceived = false;

void loop()
{
//	currentSum = 0.0;
//	currentCounter = 0;
	Serial.println("  ... loop ...");
	timeReceived = false;
//	requestTime();
//	wait(50);
	doorStatus = GetDoorStatus();
	CheckHardwareButtons(); // affects ->	buttonOpenPressed, buttonClosePressed
					            		//				buttonOpenLastStatus, buttonCloseLastStatus
						            	//				doorOpenRequested, doorCloseRequested
	MotorLoop();

	String doorString;
	switch (doorStatus)
	{
		case DoorStatus_Open:
			doorString = "Open"; break;
		case DoorStatus_InBetween:
			doorString = "InBetween"; break;
		case DoorStatus_Closed:
			doorString = "Closed"; break;
	}

	//Sensor processing
	// Collect sensor data and send through mySensors
	rht.readRHT();
	temp = rht.temperature;
	humidity = rht.humidity;
	current = ina.getCurrent_mA() + 0.7;
	voltage = ina.getBusVoltage_V();

	unsigned int data0, data1;
	boolean good;  // True if neither sensor is saturated
	if (light.getData(data0, data1))
	{
		// Geen saturatie time=0, gain=0
		// Saturatie bij time=1, gain=0 --> Lux  : 788.37 (data0 = 5047, data1 = 5047)
		// Saturatie bij time=2, gain=0 --> 
		// Saturatie bij time=0, gain=1 --> Lux  : 49.27 (data0 = 5047, data1 = 5047)

		good = light.getLux(gain, ms, data0, data1, lux);
		if (lux < 805.0F && lux > 804.0F)
			good = false;
	}

	if (luxState==0)
	{
    //+1 (wasOver) overdag, -1 (wasUnder) 's nachts
			luxState = (lux > LUX_CLOSE ? 1 : -1);
	}

	if (good) 
	{
		if (luxState<0 && lux > LUX_OPEN) {
			doorOpenRequested = true;
			luxState=1;
		}

		if (luxState>0 && lux < LUX_CLOSE) {
			//doorCloseRequested = true;
			Serial.println("Switching on the camera because of sunset - door will close shortly");
			cameraActivatedByLuxSensor = true;
			digitalWrite(Relay1Pin, HIGH);
			cameraOn = true;
			cameraStartTime = controllerTime;     
			luxState=-1;
		}
	}

	if (cameraActivatedByLuxSensor && controllerTime > cameraStartTime + DOOR_AFTER_CAMERA_TIME)
	{
		doorCloseRequested = true;
		cameraActivatedByLuxSensor = false;
		Serial.println("Door closing because of sunset");
	}

	Serial.print("Door ");
	Serial.print(doorString);
	Serial.print(", ");
	if (good) {
		Serial.print("Lux ");
		Serial.print(lux);
	}
	else
	{
		Serial.print("Lux Bad");
	}
	Serial.print(", luxState ");
	Serial.print(luxState);
	Serial.print(cameraOn ? ", camera on" : ", camera off");
	Serial.println(cameraActivatedByLuxSensor ? " (cLux)" : ", (-)");

	Serial.print("Battery ");
	Serial.print(voltage);
	Serial.print("V, ");
	Serial.print(current);
	Serial.println("mA");

	// Send sensor data to the MySensors gateway
	if (good) {
		lightMsg.set(lux, 2);
		send(lightMsg);
	}

	currentMsg.set(current, 2);
	send(currentMsg);

	voltageMsg.set(voltage, 2);
	send(voltageMsg);

	temperatureMsg.set(temp, 2);
	send(temperatureMsg);

	humidityMsg.set(humidity, 2);
	send(humidityMsg);

	int door_status_mysensors = doorStatus == DoorStatus_Closed ? 0 : 1;
	doorStatusMsg.set(door_status_mysensors);
	send(doorStatusMsg);

	byte batLvl = 0;
	if (voltage > 12.0) batLvl = 50;
	if (voltage > 12.5) batLvl = 75;
	if (voltage > 12.6) batLvl = 80;
	if (voltage > 12.7) batLvl = 90;
	if (voltage >= 12.9) batLvl = 100;
	sendBatteryLevel(batLvl);

	motorCurrentMsg.set(motorCurrent, 2);
	send(motorCurrentMsg);

	relais1Msg.set(cameraOn);
	send(relais1Msg);

	if (!timeReceived)
	{
		Serial.println("--No time received-- ");
		controllerTime += (SLEEP_PERIOD * SLEEP_CYCLES)/1000;
	}
	Serial.print("Time = ");
	Serial.print(controllerTime);
	secondsAfterMidnight = controllerTime % 86400;
	hours = secondsAfterMidnight / 3600;
	minutes = (secondsAfterMidnight % 3600) / 60;
	seconds = secondsAfterMidnight % 60;
	Serial.print(", ");
	Serial.print(hours);
	Serial.print(":");
	Serial.print(minutes);
	Serial.print(":");
	Serial.print(seconds);
	Serial.println();
  
	//End-of-loop activities
	lastDoorStatus = doorStatus;

	//Sleep for 10 seconds unless one of the hardware buttons was pressed
	sleepCounter = 0;
	while(!buttonClosePressed && !buttonOpenPressed && (sleepCounter < SLEEP_CYCLES))
	{
		//sleep(SLEEP_PERIOD);// Activate sleep (instead of wait) to decrease power usage - to be tested
		wait(SLEEP_PERIOD);  //Allows instructions from Domoticz to be sent to the node
		sleepCounter++;
		CheckHardwareButtons();
	}

	if (cameraOn && controllerTime > cameraStartTime + CAMERA_ON_TIME){
		Serial.println("Switching OFF the camera (timer)");
		digitalWrite(Relay1Pin, LOW);
		cameraOn = false;
		cameraStartTime = 0;
	}
  
}

void receiveTime(unsigned long ts) {
	Serial.print("Time received ");
	Serial.println(ts);
	controllerTime = ts;
	timeReceived = true;
}

void receive(const MyMessage &message) {
	if (message.sensor == CHILD_ID_DOOR) {
		uint8_t value = message.getByte();
		if (value == 0) {
		  doorCloseRequested = true;
      Serial.println("Door CLOSE request received");
		}
		if (value == 1) {
		  doorOpenRequested = true;
      Serial.println("Door OPEN request received");
		}
	}
	if (message.sensor == CHILD_ID_RELAIS1) {
		uint8_t value = message.getByte();
		if (value == 0) {
		  digitalWrite(Relay1Pin, LOW);
		  Serial.println("Camera switched off remotely");
		  cameraOn = false;
		  cameraStartTime = 0;
		}
		if (value == 1) {
		  digitalWrite(Relay1Pin, HIGH);
		  Serial.println("Camera switched ON remotely");
		  cameraOn = true;
		  cameraStartTime = controllerTime;
		}
	}
	if (message.sensor == CHILD_ID_RELAIS2) {
		uint8_t value = message.getByte();
		if (value == 0) digitalWrite(Relay2Pin, LOW);
		if (value == 1) digitalWrite(Relay2Pin, HIGH);
	}
}


byte GetDoorStatus()
{
	if (digitalRead(EndDetectionClosePin) == LOW)
		return DoorStatus_Closed;
	if (digitalRead(EndDetectionOpenPin) == LOW)
		return DoorStatus_Open;
	return DoorStatus_InBetween;
}

void CheckHardwareButtons()
{
	buttonOpenPressed = digitalRead(PushButtonOpenPin) == LOW;
	buttonClosePressed = digitalRead(PushbuttonClosePin) == LOW;
	//Serial.print(buttonOpenPressed ? "A" : "a");
	//Serial.print(buttonClosePressed ? "B" : "b");
	if (doorStatus != DoorStatus_Closed && ((buttonClosePressed && !buttonCloseLastStatus)))
		doorCloseRequested = true;
	if (doorStatus != DoorStatus_Open && ((buttonOpenPressed && !buttonOpenLastStatus)))
		doorOpenRequested = true;

	buttonCloseLastStatus = buttonClosePressed;
	buttonOpenLastStatus = buttonOpenPressed;
}

void MotorLoop()
{
	unsigned long motorStartTime;

	doorStatus = GetDoorStatus();
	if (doorStatus != DoorStatus_Closed && doorCloseRequested) 
	{
		ActivateMotor(MotorStatus_Close);
		doorCloseRequested = false;
	}
	if (doorStatus != DoorStatus_Open && doorOpenRequested) 
	{
		ActivateMotor(MotorStatus_Open);
		doorOpenRequested = false;
	}

	if (motorStatus != MotorStatus_Stop)
	{
		currentSum = 0.0;

		while (true)
		{
			current = ina.getCurrent_mA() + 0.7;
			currentSum += current;
			currentCounter++;
			motorCurrent = currentSum / currentCounter;
			doorStatus = GetDoorStatus();
			if (motorStatus == MotorStatus_Open && doorStatus == DoorStatus_Open) break;
			if (motorStatus == MotorStatus_Retour && doorStatus == DoorStatus_Open) break;
			if (motorStatus == MotorStatus_Close && doorStatus == DoorStatus_Closed) {
				wait(EXTRA_TIME_CLOSE);
				break;
			}
			wait(10);
			if (motorStatus != MotorStatus_Stop && millis() > motorStopTime)
			{
				Serial.println("=-=-=-=-=-= Emergency door reopening =-=-=-=-=-=");
				wait(200);
				ActivateMotor(MotorStatus_Retour);
			}
		}
		ActivateMotor(MotorStatus_Stop);
		//TODO: activatie van retour indien de open tijd overschreden werd.
	}
}

unsigned long motorStartTime;

void ActivateMotor(int newStatus)
{
	switch (newStatus)
	{
		case MotorStatus_Stop:
			digitalWrite(RelayMotorOpenPin, LOW);
			digitalWrite(RelayMotorClosePin, LOW);
			doorOpenRequested = false;
			doorCloseRequested = false;
			Serial.print(F("Motor stopped at "));
			Serial.println(millis());
			break;
		case MotorStatus_Open:
		case MotorStatus_Retour:
			digitalWrite(RelayMotorOpenPin, HIGH);
			digitalWrite(RelayMotorClosePin, LOW);
			motorStartTime = millis();
			motorStopTime = millis() + ((newStatus==MotorStatus_Open) ? MAX_LOOPTIJD : RETURN_AFTER_FAULT_TIME);
			Serial.print(F("Motor started "));
			if (newStatus == MotorStatus_Open)
				Serial.print(F("open at "));
			else
				Serial.print(F("retour at "));
			Serial.println(millis());
			break;
		case MotorStatus_Close:
			digitalWrite(RelayMotorOpenPin, LOW);
			digitalWrite(RelayMotorClosePin, HIGH);
			motorStartTime = millis();
			motorStopTime = millis() + MAX_LOOPTIJD;
			Serial.print(F("Motor started closing at "));
			Serial.println(millis());
			break;
	}
	motorStatus = newStatus;
}

void presentation()
{
	#pragma region MySensors setup and initialization
	sendSketchInfo("Kippenhok", "1.0");
  wait(50);
	// Register all sensors to gateway (they will be created as child devices)
	present(CHILD_ID_LIGHT, S_LIGHT_LEVEL);
  wait(50);
	present(CHILD_ID_MULTIMETER, S_MULTIMETER);
  wait(50);
//	present(CHILD_ID_POWER, S_POWER);
	present(CHILD_ID_TEMPERATURE, S_TEMP);
  wait(50);
	present(CHILD_ID_HUMIDITY, S_HUM);
  wait(50);
	present(CHILD_ID_DOOR, S_LOCK);
  wait(50);
	present(CHILD_ID_MOTORCURRENT, S_MULTIMETER);
  wait(50);
//	present(10, S_ARDUINO_NODE);
//  wait(50);
	present(CHILD_ID_RELAIS1, S_LIGHT);
  wait(50);
	present(CHILD_ID_RELAIS2, S_LIGHT);
	#pragma endregion
}


